#ifndef TEST1DSIMULATIONWITHDIFFERENTTIMESTEPS_HPP_
#define TEST1DSIMULATIONWITHDIFFERENTTIMESTEPS_HPP_

#include <cxxtest/TestSuite.h>
#include <vector>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include "UblasVectorInclude.hpp"
#include "TetrahedralMesh.hpp"
#include "MonodomainProblem.hpp"
#include "RegularStimulus.hpp"
#ifdef CHASTE_CVODE
#include "shannon_wang_puglisi_weber_bers_2004Cvode.hpp"
#endif
#include "AbstractCardiacCellFactory.hpp"
#include "PseudoEcgCalculator.hpp"
#include "DistributedTetrahedralMesh.hpp"
#include "Hdf5DataReader.hpp"
#include "Hdf5DataWriter.hpp"
#include "OutputFileHandler.hpp"
#include "TrianglesMeshReader.hpp"
#include "HeartConfig.hpp"
#include "AbstractCardiacProblem.hpp"
#include "PseudoEcgDataReadIn.hpp"
#include "PropagationPropertiesCalculator.hpp"
#include "GeneralIonChannelDataReadIn.hpp"
#include "PetscSetupAndFinalize.hpp"
#include "CommandLineArguments.hpp"

#ifdef CHASTE_CVODE

//Cell Factory defining cells for 1d chain.
class ShannonCardiacCellFactory : public AbstractCardiacCellFactory<1>
{
private:
	double mPeriod;
	boost::shared_ptr<RegularStimulus> mpStimulus;
public:
	ShannonCardiacCellFactory()
	: AbstractCardiacCellFactory<1>(),
	  mPeriod(2000),
	  mpStimulus(new RegularStimulus(-250000,5,mPeriod,99))
	  {
	  }

	AbstractCvodeCell* CreateCardiacCellForTissueNode(unsigned node)
	{
		AbstractCvodeCell* p_cell;

		boost::shared_ptr<AbstractIvpOdeSolver> p_solver;
		if (this->GetMesh()->GetNode(node)->GetPoint()[0] < 1e-6)
		{
			p_cell = new Cellshannon_wang_puglisi_weber_bers_2004FromCellMLCvode(p_solver, mpStimulus);
		}
		else
		{
			p_cell = new Cellshannon_wang_puglisi_weber_bers_2004FromCellMLCvode(p_solver, this->mpZeroStimulus);
		}
		p_cell->SetMinimalReset(true);

		p_cell->SetTolerances(1e-4,1e-6); //Most compounds are run with these tolerances set to 1e-4 and 1e-6. These are relaxed to 0.005 and 0.00005 for compounds with concentrations which are unable to be solved with the original tolerances
		return p_cell;
	}


	double GetPeriod()
	{
		return mPeriod;
	}
};



class Test1DSimulation : public CxxTest::TestSuite
{
private:

	// Method to apply drug block to ion channels
	void ApplyDrugBlock(double& hERG_IC50, double& hERG_hill, double& CaV_IC50, double& CaV_hill, double& NaV_IC50, double& NaV_hill, double& dose, AbstractCardiacCellInterface* cells)
	{

		const double DefaultNaVConductance = cells->GetParameter("membrane_fast_sodium_current_conductance");
		const double DefaultIkrConductance = cells->GetParameter("membrane_rapid_delayed_rectifier_potassium_current_conductance");
		const double DefaultCaLConductance = cells->GetParameter("membrane_L_type_calcium_current_conductance");

		double blockIkr =pow(1.0+pow(dose*pow(hERG_IC50,-1),hERG_hill),-1);
		double blockCaL =pow(1.0+pow(dose*pow(CaV_IC50,-1),CaV_hill),-1);
		double blockNaV = pow(1.0+pow(dose*pow(NaV_IC50,-1),NaV_hill),-1);

		cells->SetParameter("membrane_rapid_delayed_rectifier_potassium_current_conductance",DefaultIkrConductance*blockIkr);
		cells->SetParameter("membrane_L_type_calcium_current_conductance",DefaultCaLConductance*blockCaL);
		cells->SetParameter("membrane_fast_sodium_current_conductance",DefaultNaVConductance*blockNaV);

	}

public:

	void TestSolvingShannon1d()  throw (Exception)
	{
		std::string number_of_parameters = CommandLineArguments::Instance()->GetStringCorrespondingToOption("-number_of_parameters");
		std::string datafile;
		if (number_of_parameters=="2")
		{
			datafile = "PatchXpress_2Parameter_Data_Matrix.dat";
		}
		if (number_of_parameters=="1")
		{
			datafile = "PatchXpress_Data_Matrix.dat";
		}

		// Reads in ion channel data
		std::string file = "projects/Jptm2013Beattie/test/"+datafile;
		GeneralIonChannelDataReadIn data_in(file,3);
		ShannonCardiacCellFactory cell_factory;
		DistributedTetrahedralMesh<1,1> mesh;
		TrianglesMeshReader<1,1> reader("mesh/test/data/1D_0_to_1_100_elements");
		mesh.ConstructFromMeshReader(reader);

		HeartConfig::Instance()->SetVisualizeWithMeshalyzer(false);
		HeartConfig::Instance()->SetUseStateVariableInterpolation();
		HeartConfig::Instance()->SetIntracellularConductivities(Create_c_vector(1.74));

		double period = cell_factory.GetPeriod();

		double iteration;
		// Loops through compounds in data file
		for(unsigned compound_number = 0;compound_number<data_in.GetNumberOfDrugs();compound_number++)
		{
			iteration = 0;
			// Prepares monodomain equation
			MonodomainProblem<1> monodomain_problem(&cell_factory);
			monodomain_problem.SetMesh(&mesh);
			monodomain_problem.Initialise();
			std::string compound_name = data_in.GetDrugName(compound_number);
			std::vector<double> dose_vector = data_in.GetDoses(compound_number);
			std::string QTfoldername = "PatchXpress_1D_"+number_of_parameters+"_Parameter_Multiple/Compound_" + compound_name;
			OutputFileHandler results_handler2(QTfoldername);
			std::string QTResultsfilename = "QT_Results.dat";
			out_stream results_file2 = results_handler2.OpenOutputFile(QTResultsfilename);
			double hERG_IC50 = data_in.GetIC50Value(compound_number,0);
			double hERG_hill= data_in.GetHillCoefficient(compound_number,0);
			double NaV_IC50= data_in.GetIC50Value(compound_number,1);
			double NaV_hill = data_in.GetHillCoefficient(compound_number,1);
			double CaV_IC50 = data_in.GetIC50Value(compound_number,2);
			double CaV_hill = data_in.GetHillCoefficient(compound_number,2);
			std::cout<<"Starting Simulation For Compound " + compound_name<<"\n";
			// Loops through each of the doses tested for each compound
			for (unsigned dose_num = 0; dose_num < dose_vector.size(); dose_num++)
			{
				std::ostringstream ss;
				ss << std::fixed << std::setprecision(5);
				ss << dose_num;
				std::string pseudo_ecg_output_dir = "PatchXpress_1D_Results_"+number_of_parameters+"_Parameter/Compound_" + compound_name + "/Dose_" + ss.str();
				OutputFileHandler results_handler(pseudo_ecg_output_dir, true);

				// Applies drug block to cells
				for (unsigned i = 0; i<100; i++)
				{
					AbstractCardiacCellInterface* cells = monodomain_problem.GetTissue()->GetCardiacCell(i);
					ApplyDrugBlock(hERG_IC50,hERG_hill,CaV_IC50, CaV_hill, NaV_IC50, NaV_hill,dose_vector[dose_num],cells);
				}
				HeartConfig::Instance()->SetOutputDirectory(pseudo_ecg_output_dir);
				std::string QTfilename = "Compound_" + compound_name + "_QT_Results.dat";
				out_stream results_file = results_handler.OpenOutputFile(QTfilename);
				std::string pre;

				//Pacing continued until steady state is reached, that is until the difference
				//between the APD90 on successive action potentials is on successive
				//paces is less than the specified tolerance of 1e-1 for the central node.
				std::string pace_string;

				double apd90_diff;


				for (unsigned pace_num = 0; pace_num<500;pace_num ++)
				{
					std::stringstream convert;
					convert << pace_num;
					pace_string = convert.str();
					pre = "shannon_1d" + pace_string;
					HeartConfig::Instance()->SetOutputFilenamePrefix(pre);

					//Pacing continued until steady state is reached, that is until the difference
					//between the APD90 on successive action potentials is on successive
					//paces is less than the specified tolerance of 1e-1 for the central node.
					std::cout << "Solving" << endl;

					//Solving with with different pde and ode time steps over different portions of the wave

					// Section 1 - up to 90ms - just before stimulus is applied
					HeartConfig::Instance()->SetSimulationDuration(double(iteration)*period + 90);
					HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(1, 1, 1);
					monodomain_problem.Solve();

					// Section 2 - between 90 ms and 220 ms - small time steps during and just after the upstroke
					HeartConfig::Instance()->SetSimulationDuration(double(iteration)*period + 220);
					HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(0.01, 0.01, 1);
					monodomain_problem.Solve();

					// Section 3 - Between 220 and 600 ms - during repolarisation
					HeartConfig::Instance()->SetSimulationDuration(double(iteration)*period + 600);
					HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(0.1, 0.1, 1);
					monodomain_problem.Solve();

					// Section 4 - Between 600 and 2000 ms - larger time step while at resting potential
					HeartConfig::Instance()->SetSimulationDuration(double(iteration+1u)*period);
					HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(1, 1, 1);
					monodomain_problem.Solve();

					std::cout<<"Post Processing"<< std::endl;

					std::string pseudo_ecg_output_dir3 = OutputFileHandler::GetChasteTestOutputDirectory() + pseudo_ecg_output_dir;
					Hdf5DataReader simulation_data(pseudo_ecg_output_dir3,pre,false);

					// Calculating APD90 at central node to check if the solution has reached steady state
					PropagationPropertiesCalculator data(&simulation_data);
					double temp_apd90;
					double apd90 = data.CalculateActionPotentialDuration(90,50);
					apd90_diff=fabs(apd90-temp_apd90);
					temp_apd90 = apd90;
					iteration = iteration +1;

					if(apd90_diff < 1e-1)

					{
						break;
					}


				}
				// Once solution is in steady state, a final pace is performed with a much finer printing time step
				// and this solution is used to calculate the pseudo-ecg from which the QT interval is calculated
				HeartConfig::Instance()->SetOutputFilenamePrefix("shannon_1d_final");
				HeartConfig::Instance()->SetSimulationDuration(double(iteration+1u)*period);
				HeartConfig::Instance()->SetOdePdeAndPrintingTimeSteps(0.01,0.01,0.01);
				monodomain_problem.Solve();
				iteration = iteration +1;

				// Calculating pseudo-ecg
				ChastePoint<1> probe_electrode(50.0);

				std::string pseudo_ecg_output_dir2 = OutputFileHandler::GetChasteTestOutputDirectory() + pseudo_ecg_output_dir;
				PseudoEcgCalculator<1,1,1> calculator (mesh, probe_electrode, pseudo_ecg_output_dir2, "shannon_1d_final", "V",false);
				HeartConfig::Instance()->SetOutputDirectory(pseudo_ecg_output_dir);

				calculator.WritePseudoEcg();
				std::string pseudo_ecg_output_dir5 = OutputFileHandler::GetChasteTestOutputDirectory() + pseudo_ecg_output_dir;
				std::string h5file = pseudo_ecg_output_dir5 + "/shannon_1d_final.h5";
				remove(h5file.c_str());
				unsigned min_pseudo_ecg_time;
				double pseudo_ecg_min = 10000;
				std::string ECG_data = OutputFileHandler::GetChasteTestOutputDirectory() + pseudo_ecg_output_dir + "/output/PseudoEcgFromElectrodeAt_50_0_0.dat";

				// Pseudo Ecg is read back in and used to calculate the QT interval at each dose.
				// This is then written to file. The QT interval is defined as the time interval between the application of the
				// stimulus and the minimum of the T wave.

				// For most cases, the output pseudo ecg is of a standard form and the QT interval
				// is calculated by finding the minimum point of the curve after the QRS interval.
				// In a few cases, the pseudo ecg has a late, very shallow T-wave which is not always
				// the minimum encountered after the QRS interval. In these cases, the standard method
				// of measuring the QT interval gives an incorrect result (detecting the minimum to be
				//just after the QRS interval, giving a very large negative percent change in QT interval
				// as compared to the control). We identify these cases and perform an alternative measure
				// of the QT interval, by searching backwards for the first minimum encountered from the
				// end of the pseudo-ecg output.

				// Standard QT interval measurement - searching forwards for first minimum in pseudo-ecg
				// after QRS interval.
				PseudoEcgDataReadIn datain(ECG_data);
				std::vector<double> pseudo_ecg_vector (datain.GetNumberOfValues());

				min_pseudo_ecg_time = 0;
				for (unsigned i = 1; i<datain.GetNumberOfValues(); i++)
				{
					pseudo_ecg_vector[i] = datain.GetValues(i);

					if ((i>13000)&&(pseudo_ecg_vector[i]<pseudo_ecg_min))

					{
						pseudo_ecg_min = pseudo_ecg_vector[i];
						min_pseudo_ecg_time = i;
					}
				}

				double control_QT;
				double QTInterval = (0.01*min_pseudo_ecg_time) - (99);
				if (dose_vector[dose_num]==0)
				{
					control_QT = QTInterval;
				}

				double  QT_percent_change =100*(QTInterval-control_QT)*pow(control_QT,-1);

				// Alternative QT measurement performed if it is identified that the pseudo-ecg
				// may be of a non typical form and QT interval needs to be recalculated.
				if (QT_percent_change <-50)
				{
					double min_pseudo_ecg_time = 0;
					double pseudo_ecg_max = -1000;
					double min_pseudo_ecg_time_2 = 0;
					double pseudo_ecg_max_2 = -1000;
					for (unsigned i = datain.GetNumberOfValues();i>0;i--)
					{
						if ((fabs(pseudo_ecg_vector[i])>pseudo_ecg_max))

						{

							pseudo_ecg_max = fabs(pseudo_ecg_vector[i]);
							min_pseudo_ecg_time = i;


							if (fabs(pseudo_ecg_vector[i]-pseudo_ecg_vector[datain.GetNumberOfValues()])>5e-5)
							{
								for (unsigned j = min_pseudo_ecg_time;j>min_pseudo_ecg_time-8000;j--)
								{

									if ((fabs(pseudo_ecg_vector[j])>pseudo_ecg_max_2))

									{
										pseudo_ecg_max_2 = fabs(pseudo_ecg_vector[j]);
										min_pseudo_ecg_time_2 = j;

									}
								}
								goto end;
							}



						}


					}

					end:
					QTInterval= (0.01*min_pseudo_ecg_time_2) -99;
					QT_percent_change =100*(QTInterval-control_QT)*pow(control_QT,-1);
				}
				*results_file2<<dose_vector[dose_num]<<" "<<QTInterval<<" "<<QT_percent_change<<"\n";
				*results_file2<<std::flush;
				std::cout.flush();


				std::cout << " QT Interval = " << QTInterval << "\n" << std::flush;
				std::cout << " QT % Change = " << QT_percent_change << "\n" << std::flush;



			}

			results_file2->close();
		}

	}

};

#endif //CHASTE_CVODE

#endif /*TEST1DSIMULATIONWITHDIFFERENTTIMESTEPS_HPP_*/
