#ifndef TESTSINGLECELLWITHOPTIONS_HPP_
#define TESTSINGLECELLWITHOPTIONS_HPP_

#include <cxxtest/TestSuite.h>
#include <cmath>
#include <iostream>
#include <vector>
#include <string>
#include <ctime>
#include <fstream>
#include <cassert>
#include <sstream>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/shared_ptr.hpp>
#include "Exception.hpp"
#include "RegularStimulus.hpp"
#include "AbstractIvpOdeSolver.hpp"
#include "CellProperties.hpp"
#include "shannon_wang_puglisi_weber_bers_2004.hpp"
#include "shannon_wang_puglisi_weber_bers_2004Cvode.hpp"
#include "AbstractCvodeCell.hpp"
#include "AbstractCardiacCell.hpp"
#include "OdeSolution.hpp"
#include "RegularStimulus.hpp"
#include "OdeSystemInformation.hpp"
#include "VectorHelperFunctions.hpp"
#include "GeneralIonChannelDataReadIn.hpp"
#include "OutputFileHandler.hpp"
#include "SteadyState.hpp"
#include "CommandLineArguments.hpp"

class TestSingleCellWithOptions: public CxxTest::TestSuite
{
public:
	void TestSolvingSingleCell() throw(Exception)
	{

		// Code should be first compiled and then run with command line options specified. Options include:
		// assay -assay_name (PatchXpress/IonWorks/QSAR)
		// block -block_type (multiple/hERG)
		// number_of_parameter -number_of_parameters_drug_block_model_is_fitted_for (1/2) (IC50 Only or IC50 and Hill)
		// include_KCNQ1 -whether_to_include_KCNQ1_interactions_when_using_IonWorks/FLIPR_data (yes/no) (option only
		// included when using IonWorks/FLIPR data)

		std::string assay = CommandLineArguments::Instance()->GetStringCorrespondingToOption("-assay");
		std::string block = CommandLineArguments::Instance()->GetStringCorrespondingToOption("-block");
		std::string number_of_parameters = CommandLineArguments::Instance()->GetStringCorrespondingToOption("-number_of_parameters");
		std::string include_KCNQ1 = "no";

		// Only recognises option to include drug interactions with KCNQ1 channel if the data being used is from the
		// IonWorks assay - the default is set to no for other assays for which this option is not available
		if (assay == "IonWorks")
		{
			include_KCNQ1 = CommandLineArguments::Instance()->GetStringCorrespondingToOption("-include_KCNQ1");

		}

		// Data file including all the IC50 values, Hill coefficients and doses to be simulated for each compound in the correct
		// format for the code (produced by the matlab code preparingdata.m) is selected and read in according to options
		// specified when running the code
		unsigned number_of_channels=3;
		std::string datafile;

		if (assay=="PatchXpress" && number_of_parameters == "1")
		{
			datafile = "PatchXpress_Data_Matrix.dat";

		}
		if (assay=="PatchXpress" && number_of_parameters == "2")
		{
			datafile = "PatchXpress_2Parameter_Data_Matrix.dat";

		}

		if (assay=="QSAR")
		{

			datafile = "QSAR_Data_Matrix.dat";

		}

		if (assay=="IonWorks" && number_of_parameters=="1")
		{
			if (include_KCNQ1=="no")
			{
				datafile = "IonWorks_Data_Matrix.dat";
			}
			if (include_KCNQ1 == "yes")
			{
				datafile = "IonWorks_With_Iks_Data_Matrix.dat";
				number_of_channels = 4; // adjusts default of 3 channels to read in data for extra ion channel
			}
		}
		if (assay=="IonWorks" && number_of_parameters=="2")
		{
			if (include_KCNQ1=="no")
			{
				datafile = "IonWorks_2Parameter_Data_Matrix.dat";
			}
			if (include_KCNQ1 == "yes")
			{
				datafile = "IonWorks_With_Iks_2Parameter_Data_Matrix.dat";
				number_of_channels = 4; // adjusts default of 3 channels to read in data for extra ion channel
			}
		}

		// If data from the QSAR model is being used and the option to use data fitted for 2 parameters is selected
		// then this choice is overwritten as only IC50 values are available from the model (and the Hill coefficient
		// is assumed to be 1).
		if (assay == "QSAR" && number_of_parameters=="2")
		{
			number_of_parameters =1;
		}

		std::string file = "projects/Jptm2013Beattie/test/"+datafile;
		GeneralIonChannelDataReadIn data_in(file,number_of_channels);

#ifdef CHASTE_CVODE
		// Loops through compounds in the data file
		for (unsigned compound_number = 0; compound_number<data_in.GetNumberOfDrugs();compound_number++)
		{

			// Extracts ion channel data for each compound
			double freq = data_in.GetFreq(compound_number);
			double hERG_IC50 = data_in.GetIC50Value(compound_number,0);
			double hERG_hill = data_in.GetHillCoefficient(compound_number,0);
			double NaV_IC50 = data_in.GetIC50Value(compound_number,1);
			double NaV_hill = data_in.GetHillCoefficient(compound_number,1);
			double CaV_IC50 = data_in.GetIC50Value(compound_number,2);
			double CaV_hill = data_in.GetHillCoefficient(compound_number,2);
			double Iks_IC50;
			double Iks_hill;
			// Extracts KCNQ1/Iks channel data in addition when this option is specified
			if (assay=="IonWorks" && include_KCNQ1 == "yes")
			{
				Iks_IC50 = data_in.GetIC50Value(compound_number,3);
				Iks_hill = data_in.GetHillCoefficient(compound_number,3);
			}
			std::vector<double> dose_vector =  data_in.GetDoses(compound_number);
			std::string compound_name = data_in.GetDrugName(compound_number);
			std::cout<<"Starting Simulations For Compound " + compound_name<<"\n";
			//Stimulus parameters
			double stim_magnitude = -9.5; //uA/uF
			double stim_duration  = 5.0 ;  // ms
			double stim_start = 99; // ms
			double stim_period = 1000.0*pow(freq,-1); // ms

			//Defining stimulus, solver and system and associated properties
			boost::shared_ptr<RegularStimulus> p_stimulus(new RegularStimulus(stim_magnitude,stim_duration,stim_period,stim_start));
			boost::shared_ptr<AbstractIvpOdeSolver> p_solver;
			boost::shared_ptr<AbstractCvodeCell> rabbit_cvode_system(new Cellshannon_wang_puglisi_weber_bers_2004FromCellMLCvode(p_solver, p_stimulus));
			rabbit_cvode_system->SetAutoReset(false);
			rabbit_cvode_system->SetMaxSteps(1e5);
			double max_timestep = 0.001; //Change to 0.01 for faster running simulations.
			double sampling_time = max_timestep;
			rabbit_cvode_system->SetTolerances(1e-6,1e-8);
			rabbit_cvode_system->GetRelativeTolerance();
			rabbit_cvode_system->GetAbsoluteTolerance();

			//Default ion channel conductances taken directly from cellml file
			const double DefaultNaVConductance = rabbit_cvode_system->GetParameter("membrane_fast_sodium_current_conductance");
			const double DefaultIkrConductance = rabbit_cvode_system->GetParameter("membrane_rapid_delayed_rectifier_potassium_current_conductance");
			const double DefaultCaLConductance = rabbit_cvode_system->GetParameter("membrane_L_type_calcium_current_conductance");
			const double DefaultIksConductance = rabbit_cvode_system->GetParameter("membrane_slow_delayed_rectifier_potassium_current_conductance");

			//Results for each compound written to individual files
			std::string foldername = assay + number_of_parameters + "_Parameter_" + block + "/Compound_" + compound_name;

			if ((assay == "IonWorks") && (include_KCNQ1 == "yes"))
			{
				foldername = assay + "_" + number_of_parameters + "_Parameter_" + block + "_With_Iks/Compound_" + compound_name;
			}

			OutputFileHandler results_handler(foldername);
			std::string filename = "APD90_Results.dat";
			out_stream results_file = results_handler.OpenOutputFile(filename);

			double apd90;
			double apd90_percent_change;
			double control_apd90;


			//Simulations ran at each dose specified in data input file
			for (unsigned dose_index = 0; dose_index <dose_vector.size();dose_index++)
			{
				double dose = dose_vector[dose_index];
				double blockIkr=1;
				double blockCaL=1;
				double blockNaV=1;
				double blockIks=1;
				// Block of appropriate ion channels determined according to options specified when running the code
				// and the degree of block is calculated according to the IC50 value and Hill coefficient from the
				// data input file for each compound

				if ((assay=="PatchXpress"||assay=="QSAR"))

				{

					// When using data from the PatchXpress and QSAR model, a multiple channel block consists of
					// block of the hERG, NaV1.5 and CaV1.2 channels.
					if (block=="multiple")
					{
						//Simple pore block calculated and imposed on default conductances
						blockIkr =pow(1.0+pow(dose*pow(hERG_IC50,-1),hERG_hill),-1);
						blockCaL =pow(1.0+pow(dose*pow(CaV_IC50,-1),CaV_hill),-1);
						blockNaV = pow(1.0+pow(dose*pow(NaV_IC50,-1),NaV_hill),-1);

					}



					// If the option to assumed hERG channel block only is specified then only this channel is blocked.
					if (block=="hERG")
					{
						blockIkr =pow(1.0+pow(dose*pow(hERG_IC50,-1),hERG_hill),-1);
					}
				}

				if ((assay=="IonWorks"))
				{
					// When using data from the IonWorks assay, multiple channel block can include interactions
					// with the KCNQ1/Iks channel in addition to the hERG, NaV1.5 and CaV1.2 channels if this
					// option is specified

					if (block=="multiple"&&include_KCNQ1=="yes")
					{
						//Simple pore block calculated and imposed on default conductances
						blockIkr =pow(1.0+pow(dose*pow(hERG_IC50,-1),hERG_hill),-1);
						blockCaL =pow(1.0+pow(dose*pow(CaV_IC50,-1),CaV_hill),-1);
						blockNaV = pow(1.0+pow(dose*pow(NaV_IC50,-1),NaV_hill),-1);
						blockIks = pow(1.0+pow(dose*pow(Iks_IC50,-1),Iks_hill),-1);

					}
					if (block=="multiple"&&include_KCNQ1=="no")
					{
						//Simple pore block calculated and imposed on default conductances
						blockIkr =pow(1.0+pow(dose*pow(hERG_IC50,-1),hERG_hill),-1);
						blockCaL =pow(1.0+pow(dose*pow(CaV_IC50,-1),CaV_hill),-1);
						blockNaV = pow(1.0+pow(dose*pow(NaV_IC50,-1),NaV_hill),-1);

					}

					if (block=="hERG")
					{
						blockIkr =pow(1.0+pow(dose*pow(hERG_IC50,-1),hERG_hill),-1);

					}
				}

				// The maximal conductance of each of the ion channels considered is modified according to the
				// degree of drug block calculated

				rabbit_cvode_system->SetParameter("membrane_rapid_delayed_rectifier_potassium_current_conductance",DefaultIkrConductance*blockIkr);
				rabbit_cvode_system->SetParameter("membrane_L_type_calcium_current_conductance",DefaultCaLConductance*blockCaL);
				rabbit_cvode_system->SetParameter("membrane_fast_sodium_current_conductance",DefaultNaVConductance*blockNaV);
				rabbit_cvode_system->SetParameter("membrane_slow_delayed_rectifier_potassium_current_conductance",DefaultIksConductance*blockIks);

				//Runs model to steady state, returning steady state variables
				SteadyState steady_state(rabbit_cvode_system);
				steady_state.RunToSteadyState();

				N_Vector StateVarsFinal = rabbit_cvode_system->GetStateVariables();

				rabbit_cvode_system->SetStateVariables(StateVarsFinal);

				//The simulation is run one more time with the solution being printed at smaller time intervals
				double start_time = (steady_state.GetNumberOfPaces())*stim_period;
				double end_time = ((steady_state.GetNumberOfPaces())+1)*stim_period;
				OdeSolution solution = rabbit_cvode_system->Solve(start_time, end_time, max_timestep, sampling_time);

				//Post Processing of steady state action potential to determine APD90 and
				//percentage change in APD90 as compared to control
				std::vector<double> voltages = solution.GetVariableAtIndex(0);

				CellProperties ap_properties(voltages,solution.rGetTimes(),-85);

				apd90 =ap_properties.GetLastActionPotentialDuration(90);

				if (dose_index==0)
				{
					control_apd90 = apd90;
				}
				apd90_percent_change =100*(apd90-control_apd90)*pow(control_apd90,-1);

				// APD90 value and percent change in APD90 as compared to control at each dose written to file
				//for each compound
				*results_file<<dose<<" "<<apd90<<" "<<apd90_percent_change<<"\n";
				*results_file<<std::flush;

			}
			results_file->close();
		}
#endif // CHASTE_CVODE
	}
};

#endif /*TESTSINGLECELLWITHOPTIONS_HPP_*/


